from __future__ import annotations

from typing import TYPE_CHECKING, Generator, Union

if TYPE_CHECKING:
    from Akatosh import Actor


class Event:
    _at: Union[int, float]
    _priority: int
    _actor: Actor

    def __init__(self, at: Union[int, float], priority: int, actor: Actor) -> None:
        self._at = at
        self._priority = priority
        self._actor = actor

    @property
    def at(self):
        return self._at

    @property
    def priority(self):
        return self._priority

    @property
    def actor(self):
        return self._actor



from __future__ import annotations
from typing import List, Optional, Type, Union, Callable
from uuid import uuid4

from .actor import Actor


class ProductClaim:
    """Record of products claimed by a user."""

    _user: Union[Actor, object]
    _products: List

    def __init__(self, user: Union[Actor, object]) -> None:
        self._user = user
        self._products = list()

    def __eq__(self, __o: ProductClaim) -> bool:
        if self.user is __o.user:
            return True
        else:
            return False

    @property
    def user(self) -> Union[Actor, object]:
        return self._user

    @property
    def quantity(self) -> int:
        return len(self._products)

    @property
    def products(self) -> List:
        return self._products


class Producer:
    """A producer that produces products at a given rate and period."""

    _id: int
    _label: Optional[str]
    _product: Type
    _product_kargs: dict
    _inventory: List
    _capacity: Optional[int]
    _claims: List[ProductClaim]
    _production_period: Union[int, float]
    _production_rate: int
    _at: Union[int, float, Callable]
    _till: Optional[Union[int, float, Callable]]
    _priority: int

    def __init__(
        self,
        product: Type,
        production_period: Union[int, float],
        production_rate: int,
        priority: int = 1,
        capacity: Optional[int] = None,
        label: Optional[str] = None,
        at: Union[int, float, Callable] = 0,
        till: Optional[Union[int, float, Callable]] = None,
        **product_kargs,
    ) -> None:
        """Create a producer which produces products at a given rate and period.

        Args:
            product (Type): the product to be produced.
            production_period (Union[int, float]): the cycle period of production.
            production_rate (int): the amount of products to be produced in each cycle.
            priority (int, optional): the production event priority. Defaults to 1.
            capacity (Optional[int], optional): the inventory's capacity. If none, then the capacity is infinite. Defaults to None.
            label (Optional[str], optional): Label of the producer. Defaults to None.
            at (Union[int, float, Callable], optional): when the production starts. Defaults to 0.
            till (Optional[Union[int, float, Callable]], optional): when the production ends. If none, the production never ends. Defaults to None.
        """
        self._id = uuid4().int
        self._label = label or str()
        self._product = product
        self._product_kargs = product_kargs
        self._inventory = list()
        self._production_period = production_period
        self._production_rate = round(production_rate)
        self._capacity = capacity
        self._at = at
        self._till = till
        self._claims = list()

        Actor(
            at=self.at,
            step=self.production_period,
            till=self.till,
            action=self.produce,
            priority=priority,
        )

    def produce(self):
        for _ in range(self.production_rate):
            if self.capacity is None:
                self.inventory.append(self.product(**self.product_kargs))
            else:
                if len(self.inventory) < self.capacity:
                    self.inventory.append(self.product(**self.product_kargs))

    def get(self, quantity: int) -> List:
        """Get products from the inventory."""
        if quantity > len(self.inventory):
            raise ValueError(
                f"Producer {self.label} does not have enough products in inventory."
            )

        products = list()
        for _ in range(quantity):
            products.append(self.inventory.pop())

        return products

    def distribute(self, user: Union[Actor, object], quantity: int) -> bool:
        """Distribute products to a user."""
        if quantity > len(self.inventory):
            raise ValueError(
                f"Producer {self.label} does not have enough products in inventory."
            )

        for claim in self.claims:
            if claim.user is user:
                for _ in range(quantity):
                    claim.products.append(self.inventory.pop())
                return True

        claim = ProductClaim(user)
        for _ in range(quantity):
            claim.products.append(self.inventory.pop())
        self.claims.append(claim)
        return True

    @property
    def id(self) -> int:
        return self._id

    @property
    def label(self) -> Optional[str]:
        return self._label

    @property
    def product(self) -> Type:
        return self._product

    @property
    def product_kargs(self) -> dict:
        return self._product_kargs

    @property
    def inventory(self) -> List:
        return self._inventory

    @property
    def capacity(self) -> Optional[int]:
        return self._capacity

    @property
    def claims(self) -> List[ProductClaim]:
        return self._claims

    @property
    def production_period(self) -> Union[int, float]:
        return self._production_period

    @property
    def production_rate(self) -> int:
        return self._production_rate

    @property
    def at(self) -> Union[int, float, Callable]:
        return self._at

    @property
    def till(self) -> Union[int, float, Callable, None]:
        return self._till

    @property
    def num_available_products(self) -> int:
        return len(self.inventory)



from __future__ import annotations
from typing import Union, List, Optional, TYPE_CHECKING
from dataclasses import dataclass
from uuid import uuid4

if TYPE_CHECKING:
    from Akatosh import Actor


@dataclass
class ResourceClaim:
    """Resource claim record of a user. Each user only have one resource claim for a resource."""

    _user: Union[Actor, object]
    _quantity: Union[int, float]

    def __eq__(self, __o: ResourceClaim) -> bool:
        if self.user is __o.user:
            return True
        else:
            return False

    @property
    def user(self) -> Union[Actor, object]:
        return self._user

    @property
    def quantity(self) -> Union[int, float]:
        return self._quantity


class Resource:

    _id: int
    _label: str
    _capacity: Union[int, float]
    _claimed_quantity: Union[int, float]

    _claims: List[ResourceClaim]

    def __init__(
        self,
        label: Optional[str] = None,
        capacity: Union[int, float] = 1,
        init_claimed_quantity: Union[int, float] = 0,
    ) -> None:
        """Create a resource object. The resource can be claimed by users, or distribute to users. The resource can be released from users. The resource can be put back or get from the resource pool anonymously.

        Args:
            label (Optional[str], optional): the label of the resources. Defaults to None.
            capacity (Union[int, float], optional): _description_. the capacity of the resource to 1.
            init_claimed_quantity (Union[int, float], optional): the initial amount of resource in use. Defaults to 0.
        """
        self._id = uuid4().int
        self._label = label or str()
        self._capacity = capacity
        self._claimed_quantity = init_claimed_quantity

        self._claims = list()

    def get(self, quantity: Union[int, float]) -> bool:
        """Anonymous user get resource from resource pool. No resource claim is created."""
        if quantity <= self.available_quantity:
            self._claimed_quantity += quantity
            return True
        else:
            raise ValueError(
                f"Quantity {quantity} is greater than available quantity {self.available_quantity}."
            )

    def put(self, quantity: Union[int, float]) -> bool:
        """Anonymous user put resource back to resource pool. No resource claim is checked."""
        if quantity <= self.claimed_quantity:
            self._claimed_quantity -= quantity
            return True
        else:
            raise ValueError(
                f"Quantity {quantity} is greater than claimed quantity {self.claimed_quantity}."
            )

    def distribute(
        self, user: Union[Actor, object], quantity: Union[int, float]
    ) -> bool:
        """Distribute a amount of resource to a user. If the user already has a claim, the quantity will be added to the claim. Otherwise, a new claim will be created.

        Args:
            user (Union[Actor, object]): the user to distribute resource to.
            quantity (Union[int, float]): the amount of resource to distribute.

        Raises:
            ValueError: if the quantity is greater than available quantity.

        Returns:
            bool: return True if the distribution is successful.
        """
        if quantity <= self.available_quantity:
            self._claimed_quantity += quantity
            for claim in self.claims:
                if claim.user is user:
                    claim._quantity += quantity
                    return True
            resource_claim = ResourceClaim(user, quantity)
            self.claims.append(resource_claim)
            return True
        else:
            raise ValueError(
                f"Quantity {quantity} is greater than available quantity {self.available_quantity}."
            )

    def release(
        self,
        user: Optional[Actor | List[Actor] | object | List[object]] = None,
        amount: Optional[int | float] = None,
    ) -> bool:
        """Release resource from a user or group of users. If the user is no longer using any resource, the resource claim will be removed. if no user is specified, all resource claims will be removed.
        Args:
            user (Optional[Actor  |  List[Actor]  |  object  |  List[object]], optional): the user or group of users. Defaults to None.
            amount (Optional[int  |  float], optional): the amount of resource to release. Defaults to None.

        Raises:
            ValueError: raised if the amount is greater than the user claimed quantity.

        Returns:
            bool: return True if the release is successful.
        """
        if user is None:
            self._claimed_quantity = 0
            self._claims.clear()
            return True
        else:
            if isinstance(user, list):
                for u in user:
                    for claim in self.claims[:]:
                        if claim.user is u:
                            if amount is None:
                                if claim.quantity > self.claimed_quantity:
                                    raise ValueError(
                                        f"Resource claim's quantity {claim.quantity} is greater than resource's claimed quantity {self.claimed_quantity}."
                                    )
                                else:
                                    self._claimed_quantity -= claim.quantity
                                    self.claims.remove(claim)
                            else:
                                if amount > claim.quantity:
                                    raise ValueError(
                                        f"User tries to release {amount} units, but only {claim.quantity} units are claimed by the user."
                                    )
                                else:
                                    self._claimed_quantity -= amount
                                    claim._quantity -= amount
                                    if claim.quantity == 0:
                                        self.claims.remove(claim)
            else:
                for claim in self.claims[:]:
                    if claim.user is user:
                        if amount is None:
                            if claim.quantity > self.claimed_quantity:
                                raise ValueError(
                                    f"Resource claim's quantity {claim.quantity} is greater than resource's claimed quantity {self.claimed_quantity}."
                                )
                            else:
                                self._claimed_quantity -= claim.quantity
                                self.claims.remove(claim)
                        else:
                            if amount > claim.quantity:
                                raise ValueError(
                                    f"User tries to release {amount} units, but only {claim.quantity} units are claimed by the user."
                                )
                            else:
                                self._claimed_quantity -= amount
                                claim._quantity -= amount
                                if claim.quantity == 0:
                                    self.claims.remove(claim)
            return True

    @property
    def id(self) -> int:
        return self._id

    @property
    def label(self) -> str:
        return self._label

    @property
    def capacity(self) -> Union[int, float]:
        return self._capacity

    @property
    def claimed_quantity(self) -> Union[int, float]:
        return self._claimed_quantity

    @property
    def available_quantity(self) -> Union[int, float]:
        return self.capacity - self.claimed_quantity

    @property
    def claims(self) -> List[ResourceClaim]:
        return self._claims

    @property
    def users(self):
        return [claim.user for claim in self.claims]

    @property
    def utilization(self) -> Union[int, float]:
        return self.claimed_quantity / self.capacity



from Akatosh import Actor, Producer, Mundus


class Milk(Actor):
    def __init__(
        self,
        **kwargs,
    ) -> None:
        super().__init__(at=Mundus.now, step=1, priority=1, **kwargs)
        self._life_time = 0
        self._expired = False

    def action(self):
        if self.expired:
            return
        self._life_time +=1
        if self._life_time > 5:
            self._expired = True
            print(f"Milk {self.id} expired")
            if self in milkfarm.inventory:
                milkfarm.inventory.remove(self)

    @property
    def expired(self) -> bool:
        return self._expired
    
class Buyer(Actor):

    def __init__(
        self,
        **kwargs,
    ) -> None:
        super().__init__(at=0, step=1, priority=1, **kwargs)

    def action(self):
        if len(milkfarm.inventory)>=1:
            milkfarm.distribute(self, 1)
            print(f"Buyer {self.id} bought one milk")


milkfarm = Producer(product=Milk, production_period=1, production_rate=3, priority=0, capacity=10)
customer = Buyer()


Mundus.simulate(20)



from __future__ import annotations
from typing import Generator, List, Optional, Union, Callable, TYPE_CHECKING
from uuid import uuid4
from math import inf
import warnings

from Akatosh import Timeline, Mundus

if TYPE_CHECKING:
    from Akatosh import Resource
    from Akatosh import Producer


class Actor:

    _id: int
    _action: Optional[Callable]
    _timeline: Timeline
    _priority: int
    _at: Union[int, float, Callable]
    _step: Union[int, float, Callable, None]
    _till: Union[int, float, Callable, None]
    _active: bool
    _after: List[Actor]
    _followers: List[Actor]
    _label: Optional[str]

    _time: Union[int, float]
    _status: List[str]

    def __init__(
        self,
        action: Optional[Callable] = None,
        timeline: Optional[Timeline] = None,
        priority: int = 0,
        at: Union[int, float, Callable] = 0,
        step: Optional[int | float | Callable] = None,
        till: Optional[int | float | Callable] = None,
        active: bool = True,
        after: Optional[Actor | List[Actor]] = None,
        label: Optional[str] = None,
    ) -> None:

        # generate a unique id for this actor
        self._id = uuid4().int

        # assign the action to this actor
        self._action = action

        # assign the timeline to this actor
        if timeline is None:
            self._timeline = Mundus.timeline
        else:
            self._timeline = timeline
        self.timeline.actors.append(self)

        # assign the priority to this actor
        self._priority = priority

        # initialize the follower list
        self._followers = list()

        # assign the waiting target to this actor
        if after is not None:
            if isinstance(after, Actor):
                self._after = [after]
                after.followers.append(self)
                if after.priority > self.priority:
                    warnings.warn(
                        message=f"Actor {self.id} has a lower priority than its waiting target {after.id}."
                    )
            elif isinstance(after, list):
                self._after = after
                for actor in after:
                    actor.followers.append(self)
                    if actor.priority > self.priority:
                        warnings.warn(
                            message=f"Actor {self.id} has a lower priority than its waiting target {actor.id}."
                        )
            else:
                raise TypeError(f"Actor {self.id} has a wrong type of waiting target.")

        # initialize the time
        if callable(at):
            self._time = round(at(), 3)
        else:
            self._time = round(at, 3)

        if callable(step):
            self._step = step
        elif step is None:
            self._step = step
        else:
            self._step = round(step, 3)

        if callable(till):
            self._till = till
        elif till is not None:
            self._till = round(till, 3)
        else:
            self._till = till

        # initialize the status
        self._status = list()
        if active and after is None:
            self.status.append("active")
        else:
            self.status.append("inactive")
        if after is not None:
            self.status.append("onhold")

        # assign the label to this actor
        self._label = label or str()

        # schedule the actor onto timeline
        if self.onhold is False and self.active is True:
            self.timeline.schedule(self)

    def perform(self):
        # non-continuous actor
        if self.active and self.onhold is False:
            if self.step is None and self.till is None:
                self.action()
                self.status.append("completed")
                self.status.append("terminated")
                for actor in self.followers:
                    if actor.onhold:
                        actor.activate(force=False)
            # continuous actor but misses step size
            elif self.step is None and self.till is not None:
                raise AttributeError(f"Actor {self.id} has no step size defined.")
            # continuous actor with step size but no end time
            elif self.step is not None and self.till is None:
                self._till = inf
                while True:
                    self.action()
                    # if this is the last step, activate all followers
                    if self.time == self.till:
                        self.status.append("completed")
                        self.status.append("terminated")
                        for actor in self.followers:
                            if actor.onhold:
                                actor.activate(force=False)
                    if callable(self.step):
                        self._time += round(self.step(), 3)
                    else:
                        self._time += round(self.step, 3)
                    if self.time <= self.till: # type: ignore
                        yield self.timeline.schedule(self)
                    else:
                        break

            # continuous actor with step size and end time
            elif self.step is not None and self.till is not None:
                if callable(self.till):
                    self._till = round(self.till(), 3)
                while True:
                    self.action()
                    # if this is the last step, activate all followers
                    if self.time == self.till:
                        self.status.append("completed")
                        self.status.append("terminated")
                        for actor in self.followers:
                            if actor.onhold:
                                actor.activate(force=False)
                    if callable(self.step):
                        self._time += round(self.step(), 3)
                    else:
                        self._time += round(self.step, 3)
                    if self.time <= self.till: # type: ignore
                        yield self.timeline.schedule(self)
                    else:
                        break

            

    def deactivate(self, terminate: bool = True):
        # if the actor is already inactive, do nothing
        if self.inactive:
            return
        else:
            # deactivate the actor
            self.status.remove("active")
            self.status.append("inactive")
            if terminate:
                self.status.append("terminated")
            else:
                if self.time == self.till:
                    self.status.append("completed")
                    self.status.append("terminated")
                else:
                    self.status.append("onhold")
            
            # activate all followers    
            for actor in self.followers:
                if actor.onhold:
                    actor.activate(force=False)

            # remove all events associated with this actor
            for event in self.timeline.events[:]:
                if event.actor.id == self.id:
                    self.timeline.events.remove(event)

    def activate(self, force: bool = True):
        # if the actor is already active, do nothing
        if self.terminated:
            warnings.warn(message=f"Actor {self.id} is terminated and cannot be activated.")
            return
        elif self.active:
            return
        else:
            # activate the actor if force is True
            if force:
                self.status.remove("inactive")
                if self.onhold:
                    self.status.remove("onhold")
                self.status.append("active")
                self._time = self.timeline.now
                self.timeline.schedule(self)
            else:
                # activate the actor if all waiting targets are completed
                if all([x.terminated for x in self.after]) is True:
                    self.status.remove("inactive")
                    if self.onhold:
                        self.status.remove("onhold")
                    self.status.append("active")
                    self._time = self.timeline.now
                    self.timeline.schedule(self)

    def request(self, resource: Resource, amount: Union[int, float] = 1) -> bool:
        return resource.distribute(self, amount)

    def release(self, resource: Resource, amount: Optional[int | float] = None) -> bool:
        if amount is None:
            return resource.release(self)
        else:
            return resource.release(self, amount)

    def consume(self, producer: Producer, amount: int = 1) -> bool:
        return producer.distribute(self, amount)

    @property
    def id(self):
        return self._id

    @property
    def priority(self):
        return self._priority

    @property
    def action(self) -> Callable:
        if self._action is not None:
            return self._action
        else:
            raise AttributeError(f"Actor {self.id} has no action defined.")

    @property
    def timeline(self):
        return self._timeline

    @property
    def time(self):
        return self._time

    @property
    def till(self):
        return self._till

    @property
    def step(self):
        return self._step

    @property
    def status(self):
        return self._status

    @property
    def after(self):
        return self._after

    @property
    def active(self):
        return "active" in self.status

    @property
    def inactive(self):
        return "inactive" in self.status

    @property
    def onhold(self):
        return "onhold" in self.status

    @property
    def scheduled(self):
        return "scheduled" in self.status

    @property
    def completed(self):
        return "completed" in self.status

    @property
    def terminated(self):
        return "terminated" in self.status

    @property
    def followers(self):
        return self._followers



from Akatosh import Actor, Mundus

test_actor = Actor(at=0, step=1, till=5, priority=0, action=lambda: print(f"Time: {Mundus.now}\tEvent Priority: {test_actor.priority}:\tTest Actor 1 is running."))
test_actor2 = Actor(at=0, step=1, till=3, priority=0, action=lambda: print(f"Time: {Mundus.now}\tEvent Priority: {test_actor.priority}:\tTest Actor 2 is running."))

def deactive_test_actor1():
    test_actor.deactivate()
    print(f"Time: {Mundus.now}\tEvent Priority: {test_actor.priority}:\tTest Actor 1 is deactived.")

test_actor3= Actor(at=3, action=deactive_test_actor1)
test_actor4 = Actor(after=[test_actor, test_actor2],step=1, priority=-1, action=lambda: print(f"Time: {Mundus.now}\tEvent Priority: {test_actor2.priority}:\tTest Actor 4 is running."))

Mundus.simulate(5)
print(test_actor.status)
print(test_actor2.status)
print(test_actor3.status)
print(test_actor4.status)


from Akatosh import Mundus, Actor, Resource

res = Resource(capacity=100)


class User(Actor):
    def __init__(self, name: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.name = name

    def action(self):
        if Mundus.now > 10:
            res.release()
            print(
                f"resources releases all claimed units, usage {res.claimed_quantity/res.capacity}."
            )
        else:
            if res.available_quantity >= 50:
                self.request(res, 10)
                print(
                    f"{self.name} requests 10 units of resource, usage {res.claimed_quantity/res.capacity}."
                )
            if res.available_quantity <= 50:
                self.release(res)
                print(
                    f"{self.name} releases current claimed units of resource, usage {res.claimed_quantity/res.capacity}."
                )


u1 = User(name="User1", at=0, step=1, till=15, priority=0)
u2 = User(name="User2", at=0, step=1, till=15, priority=0)
Mundus.simulate(15)
